/*
 * vdserver.c
 * original code: SASANO Takayoshi
 * --- public domain, no warranty.
 */

#include <basic.h>
#include <bstdio.h>
#include <bstdlib.h>
#include <btron/btron.h>
#include <tcode.h>
#include "vdctl.h"	/*  */

#ifdef	DEBUG
#define	MESSAGE_LEVEL	5
#define	D(l, x)	if ((l) <= MESSAGE_LEVEL) printf x
#else
#define	D(l, x)	/* */
#endif

#define	SECTORSIZE	512
#define	MAPSECTOR	512
#define	CMD_TIMEOUT	-1		/* TMO_FEVR */
#define	RESP_TIMEOUT	131072

LOCAL	TC	CtlName[] = {TK_v, TK_d, TK_c, TK_t, TK_l, TK_QSTN, TK_NULL};
#define	CTLCHAR	5			/* TK_QSTNΰ */

/* ꥢ */
typedef	struct	 _workarea {
	W		blockcont;
	W		dd;
	W		fd;
	W		mapid;
} VdSvrWork;

LOCAL	VdSvrWork	Wk;


/* ޥå׽ */
LOCAL	ERR	vdcmd_map(VDCTL_HDR *hdr)
{
	ERR	err;
	W	size, start;
	B	*addr;

	start = hdr->start;
	size  = hdr->size;

	/* ɥ饤Ф׵᤬Τå */
	if (size  < 0 || size > MAPSECTOR ||
	    start < 0 || (start + size) > Wk.blockcont) {
		printf("vdcmd_map: bad request  size:%d start:%d\n",
		       size, start);
		err  = ER_PAR;
		addr = NULL;
		goto fin0;
	}

	/* ߥޥåפƤΰ */
	err = ump_rec(Wk.fd, Wk.mapid);
	if (err < ER_OK) {
		printf("vdcmd_map: ump_rec %d\n", err);
	}

	/* ׵ᤵ줿ΰޥå */
	err = map_rec(Wk.fd, start * SECTORSIZE, &addr,
		      size * SECTORSIZE, F_READ | F_WRITE | F_SYSTEM);
	if (err < ER_OK) {
		printf("vdcmd_map: map_rec %d\n", err);
		addr = NULL;
		goto fin0;
	}
	Wk.mapid = err;
	D(3, ("vdcmd_map: mapid %d @ %x\n", Wk.mapid, addr));
 fin0:
	hdr->error = err;
	hdr->addr  = addr;

	return ER_OK;
}

/* λ */
LOCAL	ERR	vdcmd_eject(VDCTL_HDR *hdr)
{
	hdr->error = ER_OK;
	hdr->addr  = NULL;

	/* det_fls()¹Υե϶شdead-lock򾷤 */

	return ER_OK;
}

/* ޥɽᥤ롼 */
LOCAL	ERR	cmdresp_loop(VOID)
{
	ERR	err;
	W	rsz;
	Bool	quit;
	VDCTL_HDR	hdr;

	quit = False;

	/* λλؼޤǥ롼פ */
	while (!quit) {
		/* ޥɼ */
		err = rea_dev(Wk.dd, 0, (UB *)&hdr, sizeof(hdr), &rsz, NULL);
		if (err < ER_OK) {
			printf("cmdresp_loop: rea_dev %d\n", err);
			goto fin0;
		}

		D(3, ("cmdresp_loop: cmd %d start %d size %d\n",
		      hdr.cmd, hdr.start, hdr.size));

		/* ޥɽ */
		switch (hdr.cmd) {
		case	VDCMD_EJECT:
			err = vdcmd_eject(&hdr);
			quit = True;
			break;
		case	VDCMD_MAP:
			err = vdcmd_map(&hdr);
			break;
		default:
			err = ER_NOSPT;
			break;
		}

		/* ޥɤν˼ԤΥ쥹ݥ󥹺 */
		if (err < ER_OK) {
			hdr.error = err;
			hdr.addr  = NULL;
		}

		D(3, ("cmdresp_loop: addr %x error %d\n",
		      hdr.addr, hdr.error));

		/* 쥹ݥ */
		err = wri_dev(Wk.dd, 0, (UB *)&hdr, sizeof(hdr), &rsz, NULL);
		if (err < ER_OK) {
			printf("cmdresp_loop: wri_dev %d\n", err);
			goto fin0;
		}
	}
 fin0:
	return err;
}

/* ᥤ */
EXPORT	WERR	main(W argc, TC *argv[])
{
	WERR	err;
	W	rsz, msz, pos;
	B	*addr;
	LINK	link;
	VDCTL_DEVINFO	vdi;

	/* Υå */
	if (argc < 3) {
		printf("usage:	vdserver [a-d] [filename]\n");
		err = ER_PAR;
		goto fin0;
	}

	/* ꥢν */
	memset(&Wk, 0, sizeof(Wk));

	/* vdctlǥХΥץ */
	CtlName[CTLCHAR] = *argv[1];
	err = opn_dev(CtlName, D_UPDATE | D_EXCL, NULL);
	if (err < ER_OK) {
		printf("vdserver: opn_dev %d\n", err);
		goto fin0;
	}
	Wk.dd = err;

	/* ۥܥ塼ѥեΥץ */
	err = get_lnk(argv[2], &link, F_NORM);
	if (err < ER_OK) {
		printf("vdserver: get_lnk %d\n", err);
		goto fin1;
	}

	err = opn_fil(&link, F_UPDATE | F_EXCL, NULL);
	if (err < ER_OK) {
		printf("vdserver: opn_fil %d\n", err);
		goto fin1;
	}
	Wk.fd = err;

	/* ե륵μ */
	err = rea_rec(Wk.fd, 0, NULL, 0, &rsz, NULL);
	if (err < ER_OK) {
		printf("vdserver: rea_rec %d\n", err);
		goto fin2;
	}
	Wk.blockcont = rsz / SECTORSIZE;

	/* 쥳ɤޥåפ */
	pos = 0;				/* ޥå׳ϥֹ */
	msz = Wk.blockcont - pos;		/* Ĥꥻ */
	if (msz > MAPSECTOR) msz = MAPSECTOR;	/*  */

	err = map_rec(Wk.fd, pos * SECTORSIZE, &addr,
		      msz * SECTORSIZE, F_READ | F_WRITE | F_SYSTEM);
	if (err < ER_OK) {
		printf("vdserver: map_rec %d\n", err);
		goto fin2;
	}
	Wk.mapid = err;
	D(3, ("vdserver: mapid %d @ %x\n", Wk.mapid, addr));

	/* ۥܥ塼Ͽ */
	vdi.diskinfo.blocksize = SECTORSIZE;
	vdi.diskinfo.blockcont = Wk.blockcont;
	vdi.diskinfo.rsv       = 0;
	vdi.diskinfo.removable = 1;
	vdi.diskinfo.protect   = 0;

	/* floppy disc mode or not */
	switch (Wk.blockcont) {
	case	1440:
		vdi.diskinfo.format = DiskFmt_2DD;
		vdi.devkind = DK_DISK_FD;
		break;
	case	2880:
		vdi.diskinfo.format = DiskFmt_2HD;
		vdi.devkind = DK_DISK_FD;
		break;
	default:
		vdi.diskinfo.format = DiskFmt_STD;
		vdi.devkind = DK_DISK_UNDEF;
		break;
	}

	vdi.mapsect  = msz;
	vdi.mapaddr  = addr;
	vdi.mappos   = pos;
	vdi.cmdtmo   = CMD_TIMEOUT;
	vdi.resptmo  = RESP_TIMEOUT;

	err = wri_dev(Wk.dd, DN_VDSETDISKINFO, (UB *)&vdi,
		      sizeof(vdi), &rsz, NULL);
	if (err < ER_OK) {
		printf("vdserver: wri_dev %d\n", err);
		goto fin3;
	}

	/* ۥܥ塼ejectޥɤޤǽλʤ */
	err = cmdresp_loop();
	if (err < ER_OK) printf("vdserver: cmdresp_loop %d\n", err);

	/* λ */
	err = wri_dev(Wk.dd, DN_VDCLEARDISKINFO, NULL, 0, &rsz, NULL);
	if (err < ER_OK) printf("vdserver: wri_dev %d\n", err);
 fin3:
	ump_rec(Wk.fd, Wk.mapid);
 fin2:
	cls_fil(Wk.fd);
 fin1:
	cls_dev(Wk.dd, 0, NULL);
 fin0:
	return err;
}
